/*
 * Goldwater Distributed Web Architecture.
 *
 * Common Domain Handling
 *
 * Copyright 1996-2002 netFluid Technology Ltd
 *
 * Goldwater is written and maintained by netFluid Technology Ltd.
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * $Revision: 1.3 $  $Date: 2003/04/10 16:18:25 $
 * --------------------------------------------------------------------------
 * $Log: gwdomain.c,v $
 * Revision 1.3  2003/04/10 16:18:25  csmith
 * Removed gwxml.h leftovers
 *
 * Revision 1.2  2003/03/29 12:21:11  csmith
 * cvs merge to 0.3.4
 *
 * Revision 1.5  2003/01/04 13:24:40  chris
 * Moved __gw_find_domain_idx_in_config from gwdomain to gwreadconf
 * as it is used in quite a few places outside libgwdom
 *
 * Revision 1.4  2002/08/22 19:08:34  chris
 * Added __gw_domain_id
 *
 * Revision 1.3  2002/06/10 17:34:47  chris
 * All phlib includes have phlib/ prefix and gwlib includes have nothing.
 *
 * Revision 1.2  2002/05/26 17:52:20  chris
 * Pulled code that searches config for domain entry given a name into a new
 * function so that it can be called externally.
 *
 * Revision 1.1  2002/05/21 21:16:09  chris
 * Common Domain Handling functionality.
 *
 * --------------------------------------------------------------------------
 */
#include <unistd.h>
#include <string.h>
#include <stdio.h>

#include <goldreq.h>
#include <mutex.h>
#include <gwlog.h>
#include <regstrings.h>
#include <gwreadconf.h>

#include <gwgbprm.h>

#include <gwdomain.h>

/* ------------------------------------------------------------------------- */

static __inline__ BBDom *
fdomain( const char *domain )
{
  BBDom *domarray = NULL;
  int    ix       = 0;

  domarray = (BBDom*)((char *)bbc + bbc->BB_dom_offset);

  for( ix = 0; ix < bbc->dommax; ix++ ) {
    if( *(domarray[ix].domain) != '\0' ) {
      if( strcmp( domarray[ix].domain, domain ) == 0 ) {
	/* Found domain */
	return &(domarray[ix]);
      }
    }
  }
  
  /* Not found */
  return NULL;
}

/* ------------------------------------------------------------------------- */
/*
 * Adds a domain to domain list and returns a ptr to the entry
 * - It checks if a domain is already present and uses the same entry if it
 *   is.
 */
BBDom *
__gw_add_domain_entry( const char *domain, const char *host, const char *ns, int sport )
{
  BBDom *domarray = NULL;
  int    ix       = 0;
  int    fix      = -1;

  domarray = (BBDom*)((char *)bbc + bbc->BB_dom_offset);

  for( ix = 0; ix < bbc->dommax; ix++ ) {
    if( *(domarray[ix].domain) != '\0' ) {
      if( strcmp( domarray[ix].domain, domain ) == 0 ) {
	/* Found domain already */
	return &(domarray[ix]);
      }
    } else {
      if( fix == -1 ) fix = ix; /* First free domain slot */
    }
  }

  if( fix == -1 ) {
    /* Error - no free slots for domain */
    gw_logf( LOG_SYSERR, "No room in domain table for domain %s", domain );
    return NULL;
  }

  /* Add new domain */
  domarray[fix].domainid = fix + 1; /* Start ID's from 1 */

  strncpy( domarray[fix].domain, domain, DOM_NM_MAX );
  
  strcpy( domarray[fix].gw_ipaddr, host );
  domarray[fix].gw_port  = sport; 

  strncpy( domarray[fix].namespace, ns, NAME_SPACE_MAX );

  domarray[fix].available = GWD_UNAVAILABLE;

  return &(domarray[fix]);
}


/* ------------------------------------------------------------------------- */
/*
 * Add a domain entry straight into BB from the GWConfig.
 */
BBDom *
__gw_domain_from_config( const char *domain )
{
  int   id        = 0;
  long  len       = 0;
  long  sport     = 0;
  char *addr      = NULL;
  char *ns        = NULL;
  char  path[128] = {0};
  BBDom *domentry = NULL;

  id =__gw_find_domain_idx_in_config( domain );

  if( id < 0 ) return NULL;

  /* Read Domain Resources
   */
  sprintf( path, "%s%06d%s", R_DOMAIN_B, id, R_DOM_ADDR_L );
  if( gw_config_read_char( path, &addr, &len ) < 0 ) addr = "";

  sprintf( path, "%s%06d%s", R_DOMAIN_B, id, R_DOM_PORT_L );
  if( gw_config_read_long( path, &sport      ) < 0 ) sport = 0;

  sprintf( path, "%s%06d%s", R_DOMAIN_B, id,  R_DOM_NSPACE_L );
  if( gw_config_read_char( path, &ns, &len   ) < 0 ) ns = "";

  domentry = __gw_add_domain_entry( domain, addr, ns, sport );

  gw_logf( LOG_INTDBG, "Added domain %s (id %ld)", domain, domentry->domainid);

  return domentry;
}

/* ------------------------------------------------------------------------- */
/*
 * Sets new state of domain, and returns old.
 */
DomAvail
__gw_domain_set_state( const char *domain, DomAvail state )
{
  BBDom *domptr = NULL;
  DomAvail ost;

  if( (domptr = fdomain( domain )) == NULL ) return Failure;

  ost = domptr->available;
  domptr->available = state;

  return ost;
}

/* ------------------------------------------------------------------------- */

RetValue
__gw_remove_domain_entry( const char *domain )
{
  BBDom *domptr = NULL;

  if( (domptr = fdomain( domain )) == NULL ) return Failure;

  domptr->domain[0] = '\0';
  domptr->available = GWD_UNAVAILABLE;

  return Success;
}

/* ------------------------------------------------------------------------- */
/* Returns the ID of the specified domain.
 * Domain ID's are always +ve
 *
 * Returns 0 if Domain is not found.
 */
int
__gw_get_domain_id( const char *domain )
{
  BBDom *domptr = NULL;

  if( (domptr = fdomain( domain )) == NULL ) return 0;

  return domptr->domainid;
}

/* ------------------------------------------------------------------------- */
/* end */

