/*
 * Goldwater Domain NetServer Control
 *
 * Copyright 2002 netFluid Technology Ltd
 *
 * Goldwater is written and maintained by netFluid Technology Ltd.
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * --------------------------------------------------------------------------
 * $Id: nsctrl.c,v 1.2 2003/03/29 12:21:11 csmith Exp $
 * --------------------------------------------------------------------------
 */
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/mman.h>

#include <phlib/xl_types.h>
#include <phlib/xl_string.h>
#include <phlib/xl_memlst.h>
#include <phlib/xl_memory.h>
#include <phlib/xl_malloc.h>

#define GW_INTERNAL_FLAGS
#include <gwmsg.h>
#include <gwneng.h>
#include <gwnevent.h>
#include <gwxlapi.h>
#include <goldreq.h>
#include <bbquery.h>
#include <gwgbprm.h>
#include <gwlog.h>
#include <gwclient.h>
#include <gwdomain.h>

#include <dbcom.h>

#include "nsctrl.h"
#include "nsreply.h"
#include "nsstate.h"
#include "version.h"

/* 
 * Const strings.... 
 */
const char SRV_DESC[] = "Goldwater netServer - (c)2002 netFluid Technology";
const char SERVICE [] = "GWNETSRV";
const char REVISION[] = VERSION;


const char CRLF    [] = "\r\n";
const char CRLFCRLF[] = "\r\n\r\n";

#define MAXSERVERS    2
#define WR_BLOCK_SIZE 4096

NetService NServers[ MAXSERVERS ];

/* Prototype functions 
 */
Action read_handler ( int, void ** );
Action event_handler( EventMsg *, void ** );

/* Private Prototypes 
 */
static RetValue accept_message  ( nsSession *sc, unsigned char *head );
static RetValue decode_header   ( nsSession *sc, unsigned char *head );
static RetValue read_data       ( nsSession *sc );
static Action   finish_message  ( nsSession *sc );
static Action   read_message_data    ( int fd, nsSession *sc );
static int      resend_req_to_service( nsSession *sc );
static int      send_req_to_service  ( nsSession *sc, long len );

/* ----------------------------------------------------------------------- */

RetValue
init_ns_network()
{
  NetService ns;

  BBDom *domarray     = NULL;

  domarray = (BBDom*)((char *)bbc + bbc->BB_dom_offset);

  gw_logf( LOG_INTDBG, "Netserver binding domain %s [%s:%d]", domarray[0].domain, domarray[0].gw_ipaddr, domarray[0].gw_port );

  ns.port          = domarray[0].gw_port;
  ns.ipaddr        = domarray[0].gw_ipaddr;
  ns.desc          = SRV_DESC;
  ns.svcidstr      = SERVICE;
  ns.read_handler  = read_handler;
  ns.event_handler = event_handler;

  if( gwne_netcom_init( NServers, 0, MAXSERVERS, WR_BLOCK_SIZE ) == Failure )
    return Failure;
  if( gwne_netcom_add_handler( &ns ) == Failure ) return Failure;
  if( gwne_netcom_init_server( 1   ) == Failure ) return Failure;

  return Success;
}

/* ----------------------------------------------------------------------- */
/*
 * Read handler is controlled by several states.
 */
Action
read_handler( int fd, void ** private )
{
  Action    ret = ENABLE_READ;
  int        n  = 0;
  nsSession *sc = (nsSession*)(*private);
  unsigned char header[DBM_PAYLOAD_IX + 1];
  
  if( *private == 0 ) return MDL_ERROR;

  switch( sc->phase ) {

  case Message:
    /* Read the message header...
	 */
    if( (n = read( fd, header, DBM_PAYLOAD_IX)) <= 0 ) return FREE_SESSION;

    gw_logf( LOG_INTDBG, "Read %d bytes from stream for REQUEST.", n );

	if( accept_message( sc, header ) == Failure ) {
	  return FREE_SESSION;
    }	

	/* State will have changed now */

    break;

  case ReadAuthInfo:
    /* Read the Authorisation data from the input stream 
	 */
    if( (n = read( fd, CSStart( sc->auth_dataCS ) + CSLen( sc->auth_dataCS ),
                       sc->auth_len               - CSLen( sc->auth_dataCS )
				 )) <= 0) return FREE_SESSION;

    gw_logf( LOG_INTDBG, "Read %d bytes from stream for AUTH.", n );

    CSLen( sc->auth_dataCS ) += n;

    if( CSLen( sc->auth_dataCS ) == sc->auth_len ) {
	  /* TODO: Check Authentication (Perhaps through an event)
	   */
      gw_logf( LOG_INTDBG, "Got auth string '%*.*s'", CSLen( sc->auth_dataCS ), CSLen( sc->auth_dataCS ), CSStart( sc->auth_dataCS ) );

      /* Initialise Write Block BEFORE entering Data phase
	   */
      if( gwp_write_block_init( &sc->ctrl ) == Failure ) {
        gw_logf( LOG_SYSWARN, "Could not initialise write block" );
	    return FREE_SESSION;
      }

	  sc->phase = Data; 
	}

  case Data: 
    ret = read_message_data( fd, sc );
    break;

  default:
    break;
  }
  
  return ret;
}

/* ----------------------------------------------------------------------- */

static RetValue
accept_message( nsSession *sc, unsigned char *head )
{
   int    magic        = 0;
   int    min_auth_len = 0;
   int    domid        = 0;

   magic = read_int32_LittleEndian( head + DBM_MAGIC_NUM_IX );

   if( magic != DBM_MAGIC_NUMBER ) {
     gw_logf( LOG_SYSINFO, "Received a non Goldwater-Domain message." );
	 return Failure;
   }

   switch( read_int32_LittleEndian( head + DBM_VERSION_IX ) ) {
   case DBM_VERSION_001:
   break;

   default:
     gw_logf( LOG_SYSINFO, "Received an incompatible Goldwater-Domain message version." );
     return Failure;
   }

   /* Check the message was for the current domain
    */
   gw_logf( LOG_INTDBG, "Message received for domain '%-*.*s'",
                        DOM_NM_MAX, DOM_NM_MAX, head + DBM_TO_IX);
   if( strncmp( head + DBM_TO_IX, _Domain, DOM_NM_MAX ) != 0 ) {
     gw_logf( LOG_SYSINFO, "Unexpected message received" );
	 return Failure;
   } 

   /* Destruct message header into a GWMsgCtrl struct */
   if( decode_header( sc, head ) == Failure ) return Failure;

   /* Check that we know about the Domain that sent us the message..
	* This is stored in sc->fromdom (got from header).
	*
	* We need to get the DOMAINID of the originator domain so that we can
	* establish a return path through DomBridge.
    */
gw_logf( LOG_INTDBG, "Received Message from Domain %s", sc->fromdom );

   if( (domid = __gw_get_domain_id( sc->fromdom ) ) < 1 ) {
     gw_logf( LOG_SYSINFO, "Received message from unknown domain %s", sc->fromdom );
	 return Failure;
   }
   sc->fromdom_id = domid;


   /* Now set up authorasation processing.  First we need to read the
    * authentication data, which is done by a new READ STATE.
	*/

   /* Create memory buffer for authorisation data
    */
   switch( sc->auth_type ) {
   case DBM_AUTH_NAME: min_auth_len = DOM_NM_MAX; break;
   case DBM_AUTH_MD5 : min_auth_len = 16;         break;  /* 128 bits */
   case DBM_AUTH_X509: min_auth_len = 128;        break;  /* SET THIS !!! */
   }
   if( sc->auth_len < min_auth_len ) {
     gw_logf( LOG_SYSINFO, "Error occured during authentication." );
     return Failure;
   }

   CreateCString( &sc->auth_dataCS, sc->auth_len );

   /* Setup other CTRL parameters required
    */
   sc->ctrl.spool_dir = _SpoolDir;
   sc->ctrl.bsize     = _MaxMsgTx;
   
   /* Enter AuthRead phase
    */
   sc->phase = ReadAuthInfo;

   return Success;

}

/* ----------------------------------------------------------------------- */

static RetValue
read_data( nsSession *sc )
{
  /* Somehow, replies returned from the service that we're about to invoke
   * must be sent with a Message Type of GWM_DOM_REPLY
   */


  /* Read message data 
   * - Done by entering another phase and reading a block at a time.
   */
   
   /* Once all data has been read, Decide what message type we've received 
    * and process accordingly:
    *
	* a) If a Request, then try and invoke the requested service
	*    (making sure it is LOCAL and not at another domain)
	*    [Might want to make provision for 'forward/multi-hop' messages]
	*    If service cannot be called, set ACK to Failure (with reasons)
	*    The writing of the message must not block [Enter a WriteRequest phase
	*    which writes the message with ENOBLOCK.  If msgsnd returns with 
	*    EWBLOCK, then register an event to try the msgsnd again.
	*    Each time we try the msgsnd, we should check the expiry timeout for
	*    the message and give up once this has past (unless GW_NTIMEOUT)].
	*
	* a1) Sending the data... Need to do timeouts - done via timed events
	*
	* b) If a Reply, match up the MSGID with the 'pending' set in SHM (created
	*    by the domain bridge when it sent the 'request' to the remote domain)
	*    inorder to obtain the reply msgqueue that the reply should be dropped
	*    into.
	*    When scanning the pending set, note should be taken of the expiry
	*    timeouts of each entry, cleaning them if they're over budget (unless
	*    GW_NTIMEOUT is set).
	*
	*    If the received reply does not match any in the pending set, silently
	*    drop the message (or INTDBG it) as the 'client' timing out would have
	*    logged the timeout condition already.
	*/

    /* Sending message, notes: 
	 *
	 * - query_bb for service and return failure to remote domain if we need to
	 * - Make sure service is NOT remote (check sinfo flags)
	 * - Find request queue for service by looking in bbq (listen_ident = 1)
	 * - Add function name and domainid from service info (query_bb)
	 * - __gw_increment_queue_count for request queue
	 * - Set message type to be GWM_SVC_REQUEST
	 * - Set TimedEvent (_ReqTout) [Set this BEFORE even poll is entered]
	 *
	 * - Send Event Poll:
	 * - gwp_write_block_data( &ctrl, data, len ) to send data block
	 * - gwp_write_block_data( &ctrl, "", 0 ) to send data already buffered
	 * - gwp_write_block_flush( &ctrl ); 
	 *   (Cleanup if any of the writes above fail)
	 * - Cancel TimedEvent
	 */
  return Success;
}


/* ----------------------------------------------------------------------- */

Action
event_handler( EventMsg *msg,  void ** private )
{
  Action  ret  = FINISHED;
  int     u_ev;
  nsSession *sc = (nsSession*)(*private);
  
  switch( msg->event ) {
    
  case EV_NEW_SESSION: /* Called when a new session is started */
    {
      nsSession  *ns;

      /* Create session memory */ 
      if( (ns=(nsSession*)xl_malloc(sizeof(nsSession)) ) != NULL ) {
 	
	    ns->phase = Message;

	    *private  = (void*)ns;  /* And now *private is the ptr to our Session*/

        gw_logf( LOG_INTDBG, "NetServer START NEW SESSION" );
	    ret = ENABLE_READ;
      }
    }
    break;
  
  case EV_END_SESSION:
    /* Invoked by the engine to clean up and exit a session.  
	 * NO WRITES or events must be made we effectively no longer have a 
	 * connection to the remote client/server.
     * 
     * Private space must be cleared here - Hence no further processing for
     * this session is possible once this function is exited.
     */
    free_cstring( &sc->auth_dataCS );

    /* unmap spool file
     */
    if( sc->ctrl.cflags & GWC_SPOOL ) {
      if( sc->ctrl.spfileptr != NULL ) {
        munmap( sc->ctrl.spfileptr, sc->ctrl.total_size );
        sc->ctrl.spfileptr = NULL;
      }
      unlink( sc->ctrl.spoolfile );
      sc->ctrl.cflags &= ~GWC_SPOOL;
    }

    gw_logf( LOG_INTDBG, "NetServer END SESSION" );

    xl_free( sc ); /* Drop session memory */
    ret = TERMINATE_SESSION;    /* WE MUST RESPOND WITH THIS!  MANDATORY! */
    break;
    
  case EV_WRDONE:
    break;

  case EV_TIME:
    u_ev = msg->user_data; /*Get user assigned event code (from EV_USER_BASE)*/
    
    switch( u_ev ) {
      case EUT_ACK_TIMEOUT:
        sc->phase = Message;
        break;

      case EU_REQ_TIMEOUT:  /* Request Timeout Event */
        /* Remove some events, just in case */
        xlapi_remove_event( EV_USER_EVENT, EU_REQMSGSND_EAGAIN, 0 );
 
        sc->phase = Message;
        break;
    }
    break;

    
  case EV_USER_EVENT:
    u_ev = msg->user_data;/*Get user assigned event code (from EV_USER_BASE)*/
    
    switch( u_ev ) {
	/* When we want to end our session NICELY we register EU_MY_SESSION_QUIT
	 * This will gracefully wait for all writes to complete and perform
	 * the tidy-ups that are required.
	 */
	case EU_MY_SESSION_QUIT: /* Because of this code - DO NOT do a write API! */
      if( xlapi_write_done() == 0 ) {
	    /* Here wait for all critical tidyups to be done.... */
        ret = FREE_SESSION; /* Indicate we want to get out */
      } else {
	    xlapi_register_event( EV_USER_EVENT, EU_MY_SESSION_QUIT );
      }
      break;

    case EU_REQMSGSND_EAGAIN: /* Try and send Request message again */
	  if( resend_req_to_service( sc ) == 1 ) {
	    if( sc->dlen <= 0 ) finish_message( sc ); /* Only if msg was sent */
	  }
	  break;

    case EU_REPMSGSND_EAGAIN: /* Try and send Reply message again */
	  if( resend_reply_to_client( sc ) == 1 ) {
	    if( sc->dlen <= 0 ) finish_message( sc ); /* Only if msg was sent */
	  }
	  break;

    case EU_MSG_PENDING:
      break;
    }
  }
  
  return ret;
}

/* ----------------------------------------------------------------------- */

static RetValue
decode_header( nsSession *sc, unsigned char *head )
{
  sc->ctrl.msgtype    = read_int32_LittleEndian( head + DBM_MSG_TYPE_IX );

  strncpy( sc->service, head + DBM_SERVICE_IX, SVC_NM_MAX );
  strncpy( sc->fromdom, head + DBM_FROM_IX,    DOM_NM_MAX );

  sc->ctrl.flags       = read_int32_LittleEndian( head + DBM_FLAGS_IX );
  sc->ctrl.total_size  = read_int32_LittleEndian( head + DBM_DATA_LEN_IX );
  sc->ctrl.org_r_msgqid= read_int32_LittleEndian( head + DBM_MSG_REP_ID_IX );
  sc->reqtout          = read_int32_LittleEndian( head + DBM_TIMEOUT_IX );

  memcpy( (void*)sc->ctrl.msgid, (void*)(head + DBM_MESSAGE_ID_IX), sizeof(uuid_t) );
  memcpy( (void*)sc->txid,  (void*)(head + DBM_TXID_IX),       sizeof(uuid_t) );

  sc->auth_len        = read_int32_LittleEndian( head + DBM_AUTH_LEN_IX  );
  sc->auth_type       = read_int32_LittleEndian( head + DBM_AUTH_TYPE_IX );

  sc->dlen            = sc->ctrl.total_size;

  /* Sort out some flags so we do the right thing with the message block to
   * be formed (just to make sure and not to rely on the remote sender!)
   */
  switch( sc->ctrl.msgtype ) {
  case GWM_DOM_REQUEST:
    sc->ctrl.flags  &= ~GWP_REPLY;   /* Clear reply flag */
	sc->ctrl.cflags |=  GWC_FORCERM; /* Force unlinking of spoolfiles */
  break;

  case GWM_DOM_REPLY:
    sc->ctrl.flags  |= GWP_REPLY;   /* Set reply flag */
	sc->ctrl.cflags |= GWC_FORCERM; /* Force unlinking of spoolfiles */
  break;

  default:
  break;
  }

  return Success;
}

/* ----------------------------------------------------------------------- */
/*
 * Sends request to service via message queue.
 * - Request may contain either the full data set or a spoolfile filename
 *
 * Return:
 *  1 = Successfully sent to service
 *  0 = Blocked, will try again (via events)
 * -1 = Failure to send to service
 */
static int
send_req_to_service( nsSession *sc, long len )
{
  BBReqQueue *bbq    = NULL;
  BBSvc *sinfo       = NULL;
  BBDom *dinfo       = NULL;
  static pid_t mypid = -1;

  if( mypid < 0 ) mypid = getpid();
  
  /* Lookup service and transfer service info block into session */

  if(  query_bb( sc->service, &sinfo, &dinfo, 
                _ServiceList, _ServClass, _ServId, bbc->domainid) == Failure ) {
    gw_logf( LOG_SYSWARN, "Cannot find service '%s'", sc->service );

    send_ack( sc, GWDomFailure, GWE_NOSVC );
    return -11;
  }
  memcpy( &(sc->svc_info), sinfo, sizeof(BBSvc) );

  /* Get reference to BBReqQueue array
   */
  bbq = (BBReqQueue*)((char*)bbc + bbc->BB_queue_offset);

  sc->req_msgf.ident       = 1;
  sc->req_msgf.dsize       = len;
  sc->req_msgf.flags       = sc->ctrl.flags;
  sc->req_msgf.total_size  = sc->ctrl.total_size;   /* Total data len */
  sc->req_msgf.dest_srv_id = sc->ctrl.destsrvid;
  sc->req_msgf.dest_dom_id = sc->fromdom_id;        /* ORIGINATORS DOMAIN ID */
  sc->req_msgf.rep_msgqid  = bbq[ bbc->dbridge_msgq_idx ].msgqid;
  sc->req_msgf.org_rep_msgqid = sc->ctrl.org_r_msgqid;
  sc->req_msgf.msgtype     = GWM_SVC_REQUEST;
  uuid_copy( sc->req_msgf.msgid, sc->ctrl.msgid );
  sc->req_msgf.clid        = mypid;
  
  /* Service (or alias) name */
  strncpy( sc->req_msgf.service, sc->svc_info.service, SVC_NM_MAX );
  sc->req_msgf.service[ SVC_NM_MAX ] = '\0';
  
  /* Function name */
  strncpy( sc->req_msgf.function, sc->svc_info.function, SVC_NM_MAX );
  sc->req_msgf.function[ SVC_NM_MAX ] = '\0';
  
  if( !(sc->req_msgf.function[0]) ) { /* Eeek - No function! */
    gw_logf( LOG_SYSERR, "EEk no function" );
    send_ack( sc, GWDomFailure, GWE_NOSVC );
	return -11;
  }


  return resend_req_to_service( sc );
}

/* ----------------------------------------------------------------------- */
/*
 * Return:
 *  1 = Successfully sent to service
 *  0 = Blocked, will try again (via events)
 * -1 = Failure to send to service
 */
static int
resend_req_to_service( nsSession *sc )
{
  long mlen       = 0;
  BBReqQueue *bbq = NULL;
  
  /* Get reference to BBReqQueue array */
  bbq = (BBReqQueue*)((char*)bbc + bbc->BB_queue_offset);

  mlen = ((long)&sc->req_msgf.data - (long)&sc->req_msgf.flags) + sc->req_msgf.dsize;
  
  gw_logf( LOG_INTDBG, "Sending data to qid %ld", 
           bbq[ sc->svc_info.req_queue_idx ].msgqid );

  if( msgsnd( bbq[ sc->svc_info.req_queue_idx ].msgqid,
	      (unsigned char*)&(sc->req_msgf), mlen, IPC_NOWAIT ) < 0 ) {
    
    if( errno == EAGAIN ) {
      /* msgsnd would have blocked..... Register event to re-send
       */
      xlapi_register_event( EV_USER_EVENT, EU_REQMSGSND_EAGAIN );
      return 0;
    }

    gw_logf( LOG_SERIOUS, "msgsnd failure (error %ld)", (long)errno );
    send_ack( sc, GWDomFailure, GWE_SYSFAIL );
    return -1;
  }

  __gw_increment_queue_count( sc->svc_info.req_queue_idx );

  send_ack( sc, GWSuccess, 0 );
  return 1;
}

/* ----------------------------------------------------------------------- */
/*
 * Reads request data into msg block ready for sending to the relevent server,
 * or straight into a spoolfile if the data is significantly large (in which 
 * case, when all data has been spooled, the filename is sent to the server
 * via a message).
 *
 * **** IF DATA WON'T FIT WITHIN ONE MSG, IT IS SENT AS A SPOOLED FILE ****
 *
 * If all was successful and the request did not specify NO_REPLY, then the
 * session is put into the result phase.
 */
static Action
read_message_data( int fd, nsSession *sc )
{
  Action  ret  = FINISHED;
  char   *rptr = NULL;
  int     n    = 0;
  int     sret = 0;
  int     b2rd = 0;

  /* Read data straight into message block
   */
  if( sc->ctrl.cflags & GWC_SPOOL ) { /* Write to spool file */
    rptr = sc->ctrl.spfileptr + sc->ctrl.bytes_spooled;
    b2rd = ( sc->dlen >= WR_BLOCK_SIZE ) ? WR_BLOCK_SIZE : sc->dlen;
    
  } else { /* Write to message block (msg size <= _MaxMsgTx) */

    /* Quick paranoid sanity check */
    if( sc->ctrl.bytes_spooled == 0 && sc->dlen > _MaxMsgTx ) {
      gw_logf( LOG_SERIOUS, "Data exceeds MSGMNB" );
      return FREE_SESSION;
    }

    rptr = (uchar*)sc->req_msgf.data + sc->ctrl.bytes_spooled;
    b2rd = sc->dlen;
  }
  
  /* read data into required buffer
   */
  n = read( fd, rptr, b2rd );

  gw_logf( LOG_INTDBG, "Read %d bytes of data stream (wanted %d)", n, b2rd );
  if( n <= 0 ) return FREE_SESSION; /* Connection Probably Died */
  
  gw_logf( LOG_INTDBG, "Data Phase: Read %d bytes.  (total_len is %ld)", 
	   n, sc->ctrl.total_size );
  
  sc->dlen               -= n;
  sc->ctrl.bytes_spooled += n;
  
  /* If we've got all the data, then finish up */
  if( sc->dlen <= 0 ) {
    
    /* xlapi_register_time_event( EU_REQ_TIMEOUT, _ReqTout ); */
    
    if( sc->ctrl.cflags & GWC_SPOOL ) {
      /* We're spooling, so finish off by sending message to service giving
       * it the spool filename, and unmap the spoolfile.
       */
      strcpy( (char*)&(sc->req_msgf.data), sc->ctrl.spoolfile );

      /* Do as appropriate - Request/Reply message
	   */
      switch( sc->ctrl.msgtype ) {
	    case GWM_DOM_REQUEST:
          if( (sret = send_req_to_service( sc, 
				           strlen(sc->ctrl.spoolfile)+1)) == -1 ) {
	        ret = FREE_SESSION;
          }
		  break;
  
        case GWM_DOM_REPLY:
          if( (sret = send_reply_to_client( sc, 
				           strlen(sc->ctrl.spoolfile)+1)) == -1 ) {
	        ret = FREE_SESSION;
          }
	      break;

         default:
		 break;
        }

    } else {
      /* We're not a spoolfile, and have all our data in a single msg blk 
	   * so the message we send contains the data we're sending, so the
	   * msg we use here MUST BE THE DATA LENGTH (compare this with the
	   * msg sent when we use a spoolfile....)
	   */
      switch( sc->ctrl.msgtype ) {
	    case GWM_DOM_REQUEST:
          if( (sret = send_req_to_service( sc, sc->ctrl.total_size )) == -1 ) 
		     ret = FREE_SESSION;
		  break;

        case GWM_DOM_REPLY:
          if( (sret = send_reply_to_client( sc, sc->ctrl.total_size )) == -1 ) 
		     ret = FREE_SESSION;
		  break;

        default:
		break;
      }
    }
  
    if( sret != 0 ) { /* Only if msg was sent or an error occurred */
	  /* Send Ack and change phase */
	  finish_message( sc ); 
    }
  }

  return ret;
}

/* ----------------------------------------------------------------------- */
/*
 * Switch states when request has been sent successfully.
 */
static Action
finish_message( nsSession *sc )
{
  Action ret = FINISHED;
  static char abuf[ DBR_PAYLOAD_IX ];

  gw_logf( LOG_INTDBG, "Status of call: %s", sc->status.result == GWSuccess ? "PASS":"FAIL" );

  store_int32_LittleEndian( abuf + DBR_STATUS_IX,   (int32)sc->status.result );
  store_int32_LittleEndian( abuf + DBR_SYS_ERROR_IX,(int32)sc->status.error  );

  xlapi_write_data( abuf, DBR_PAYLOAD_IX );
  
  /* Unmap spool file 
   */
  if( sc->ctrl.cflags & GWC_SPOOL ) {
    if( sc->ctrl.spfileptr != NULL ) {
      munmap( sc->ctrl.spfileptr, sc->ctrl.total_size );
      sc->ctrl.spfileptr = NULL;
    }
    
    /* If we sent the spoolfile reference, don't unlink, else ditch it.
	 */	
	if( sc->status.result != GWSuccess ) {
	  unlink( sc->ctrl.spoolfile );
    }

    sc->ctrl.cflags &= ~GWC_SPOOL;
  }

  sc->phase = Message;
  ret = ENABLE_READ;  /* We should never have diabled read anyway... */
    
  return ret;
}

/* ----------------------------------------------------------------------- */

void
send_ack( nsSession *sc, GWStatus stat, int err )
{
  sc->status.result = stat;
  sc->status.error  = err;
}

/* ----------------------------------------------------------------------- */
/* end */
